/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

//`define SUPPORT_SIO_COMATIBLE

//
// Module Name: SN74281
//   ver 001
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                           ALU_WIDTH:   4         8
// Number of Slice Flip Flops:            4         8
// Number of 4 input LUTs:               79       115
// Number of occupied Slices:            40        60
// Total Number of 4 input LUTs:         79       116
// Number of bonded IOBs:                22        30

// Post-PAR Static Timing Report
// Maximum frequency(MHz):           144.238    97.466
//
module SN74382_ACC # (
	parameter ALU_WIDTH = 8
) (
		 input   C_IN       // Carry Input
		, input  [ALU_WIDTH-1:0] A     // DATA IN
		, output [ALU_WIDTH-1:0] F   // DATA OUT
		, output  C_OUT      // carry-look-ahead
		, output  OVF        // overflow

		, input   [2:0] S    // Function Select
		, input   [1:0] RS   // Reg. Select
		, input   RC         // Reg. Control
		, input   CP         // Clock 
`ifdef SUPPORT_SIO_COMATIBLE
		, inout   SIO0        // RI/LO
		, inout   SIO3        // LI/RO
`else
		, input   SI0        // RI/LO
		, output  SO0        // RI/LO
		, input   SI3        // LI/RO
		, output  SO3        // LI/RO
`endif
		);
	wire [ALU_WIDTH-1:0] B;

	alu_74382 # ( .ALU_WIDTH(ALU_WIDTH) ) ALU (
		  .C_IN(C_IN)     // Carry Input
		, .A(A[ALU_WIDTH-1:0])      // DATA IN
		, .B(B[ALU_WIDTH-1:0])      // DATA IN
		, .F(F[ALU_WIDTH-1:0])      // DATA OUT
		, .C_OUT(C_OUT)   // carry-look-ahead
		, .OVF(OVF)       // overflow
		, .S(S[2:0])      // Function Select
		);
	acc_74281 # ( .ALU_WIDTH(ALU_WIDTH)) ACC (
		  .B(B[ALU_WIDTH-1:0])      // DATA OUT
		, .F(F[ALU_WIDTH-1:0])      // DATA IN

		, .RS(RS[1:0])    // Reg. Select
		, .RC(RC)         // Reg. Control
		, .CP(CP)         // Clock 
		, .SI0(SI0)       // RI/LO
		, .SO0(SO0)       // RI/LO
		, .SI3(SI3)       // LI/RO
		, .SO3(SO3)       // LI/RO
		);

`ifdef SUPPORT_SIO_COMATIBLE
	assign SIO0 = (RS == 2) ? SO0 : "Z" ;
	assign SIO3 = (RS == 1) ? SO3 : "Z" ;

	assign SI0 = SIO0;
	assign SI3 = SIO3;
`endif

endmodule

module alu_74382 # (
	parameter ALU_WIDTH = 4
) (
		  input   C_IN       // Carry Input
		, input  [ALU_WIDTH-1:0] A    // DATA IN
		, input  [ALU_WIDTH-1:0] B    // DATA IN
		, output [ALU_WIDTH-1:0] F    // DATA OUT
		, output  C_OUT      // carry-look-ahead
		, output  OVF        // overflow

		, input  [2:0] S   // Function Select
		);

    wire [ALU_WIDTH-2:0] minus1 = -1;

`define ZZZ
`ifdef ZZZ
    wire [ALU_WIDTH:0] Rd_out =
		  (S == 1) ? B - A - C_IN
		: (S == 2) ? A - B - C_IN
		: (S == 3) ? A + B + C_IN
		: (S == 4) ? { 1'b0,  (A ^ B) } 
		: (S == 5) ? { 1'b0,  (A | B) }
		: (S == 6) ? { 1'b0,  (A & B) }
		: (S == 7) ? { 1'b0, minus1 , C_IN }
		: 0;
`else
    wire [ALU_WIDTH-1:0] Rd_in = (S == 1) ? B : A;
    wire [ALU_WIDTH-1:0] Rr_in = (S == 1) ? A : B;

    wire [ALU_WIDTH:0] Rd_out = 
		 ((S == 1) | (S == 2)) ? Rd_in  - Rd_in - C_IN
		: (S == 3) ? Rd_in + Rr_in + C_IN
		: (S == 4) ? { 1'b0,  (Rd_in ^ Rr_in) } 
		: (S == 5) ? { 1'b0,  (Rd_in | Rr_in) }
		: (S == 6) ? { 1'b0,  (Rd_in & Rr_in) }
		: (S == 7) ? { 1'b0, minus1 , C_IN }
		: 0;

   assign OVF = (S == 3) ? (Rd_out[ALU_WIDTH-1] 
				& ~Rd_in[ALU_WIDTH-1] & ~Rr_in[ALU_WIDTH-1]
                	    | ~Rd_out[ALU_WIDTH-1] 
				&  Rd_in[ALU_WIDTH-1] &  Rr_in[ALU_WIDTH-1] )
	      : ( (S == 1) | (S == 2) )
			? (Rd_out[ALU_WIDTH-1]
				 & ~Rd_in[ALU_WIDTH-1] &  Rr_in[ALU_WIDTH-1]
               			| ~Rd_out[ALU_WIDTH-1]
				 &  Rd_in[ALU_WIDTH-1] & ~Rr_in[ALU_WIDTH-1] )
	      : 0 ;
`endif

    assign F     = Rd_out[ALU_WIDTH-1:0];
    assign C_OUT = Rd_out[ALU_WIDTH];

endmodule

module acc_74281 # (
	parameter ALU_WIDTH = 4
) (
		  output [ALU_WIDTH-1:0] B    // DATA OUT
		, input  [ALU_WIDTH-1:0] F    // DATA IN

		, input  [1:0] RS   // Reg. Select
		, input   RC         // Reg. Control
		, input   CP         // Clock 
		, input   SI0        // RI/LO
		, output  SO0        // RI/LO
		, input   SI3        // LI/RO
		, output  SO3        // LI/RO
		);

    reg [ALU_WIDTH-1:0] acc;

    assign B = acc; 
    assign SO0 =  F[0];
    assign SO3 = RC ? F[ALU_WIDTH-2] : F[ALU_WIDTH-1];

    always @(posedge CP)
    begin
	acc  <=   (RS == 0) ? F
		: (RS == 1) ? { ( RC ? acc[ALU_WIDTH-1] : F[ALU_WIDTH-2] )
				, F[ALU_WIDTH-3:0] , SI0 }
		: (RS == 2) ? { ( RC ? acc[ALU_WIDTH-1] : SI3 )
				, ( RC ? SI3 : F[ALU_WIDTH-1] )
				, F[ALU_WIDTH-2:1] }
		: acc;
    end

    initial
    begin
	acc  <= 0;
    end

endmodule
