/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

`ifdef USE_ALU_MODULE

//`define ALU_BY_LUT
//`define ZFLAG_TUNE_1

  `ifdef ALU_BY_LUT
  `else
    `define ALU_BY_LOGIC
  `endif

// global defines :  USE_ALU_MODULE , ALU_BY_LUT

//
// Module Name: rtavr_alu
//   ver 001
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                        all-in-one  
// Number of Slice Flip Flops:              xxx
// Number of 4 input LUTs:                  xxx
// Number of occupied Slices:               xxx
// Total Number of 4 input LUTs:            xxx
// Number of bonded IOBs :                  xxx
//   IOB Flip Flops :                       xxx
// Number of BUFGMUXs :                     xxx
// Number used for Dual Port RAMs           xxx
// Number of RAMB16BWEs :                   xxx
// Clock to Setup 

module rtavr_alu (
		  input [7:0] A
		, input [7:0] B
		, output [7:0] F

		, input C_in
		, input H_in
		, input V_in
		, input Z_in
		, output C_out
		, output H_out
		, output V_out
		, output Z_out

		, input [2:0] S
		);

// S == 0 : ADD (wo carry)
// S == 1 : ADC (with carry)
// S == 2 : SBC (with carry)
// S == 3 : SUB (wo carry)
// S == 4 : AND
// S == 5 : XOR
// S == 6 : OR
// S == 7 : MOV
//
//                      S = 0  1  2  3  4  5  6  7
// C_out  : Carry[8]        o  o  o  o  -  -  -  -
// H_out  : Carry[4]        o  o  o  o  -  -  -  -
// V_out  : Over Flow       o  o  o  o  -  -  -  -
// Z_out  : Zero            o  o  o (o) -  -  -  -
//          (o) if (Z_in == 1)
//
// OP3          --*-:**------------                               HSVNZC
// CP/CPC       000B:01rd:dddd:rrrr	# B=1: CP  B=0: CPC   #   XXXXXX
// SUB/SBC	000B:10rd:dddd:rrrr	# B=1: sub B=0: sbc   #   XXXXXX
// ADD/ADD	000C:11rd:dddd:rrrr	# C=0: add C=1: adc   #   XXXXXX
// AND		0010:00rd:dddd:rrrr	# logical AND         #   -X0XX-
// EOR		0010:01rd:dddd:rrrr	# exclusive OR        #   -X0XX-
// OR		0010:10rd:dddd:rrrr	# inclusive OR        #   -X0XX-
// MOV		0010:11rd:dddd:rrrr	# copy register       #   ------

`ifdef ALU_BY_LOGIC
     wire [8:0] ALU_OUT = ((S == 0) | (S == 1)) ? ({1'b0, A}  + {1'b0, B }
				   + {8'b0, (S[0])? C_in : 1'b0 })
		     : ((S == 2) | (S == 3)) ? ({1'b0,A} - {1'b0, B }
				   - {8'b0, (S[0]) ? 1'b0: C_in })
		     : (S == 4) ? {C_in, (A & B) }
		     : (S == 5) ? {C_in, (A ^ B) }
		     : (S == 6) ?  {C_in, (A | B) }
		     : {C_in, B } ;

     assign F = ALU_OUT[7:0];

     assign C_out = ALU_OUT[8]; 

     assign Z_out = (S == 7) & Z_in
		  | ( (S == 2) & Z_in | ~((S == 2)|(S == 7)) )
			 & (ALU_OUT[7:0] == 0);

     assign V_out = ((S == 0)|(S == 1)) & (F[7] & ~A[7] & ~B[7]
                		| ~F[7] &  A[7] &  B[7] )
	      | ((S == 2)|(S == 3))  & (   F[7] & ~A[7] &  B[7]
               			| ~F[7] &  A[7] & ~B[7] )
	      | (S == 7) & V_in ;
	      //  ((S == 4)|(S == 5)|(S == 6)) ? 0

     assign H_out = ((S == 0) | (S == 1)) & ( 
			  ~F[3] & A[3] 
			| ~F[3] & B[3]
			|  A[3]  & B[3] )
	       | ((S == 2) | (S == 3)) & (
			   F[3] & ~A[3] 
			|  F[3] &  B[3]
			| ~A[3]  &  B[3] )
	       | (S[2]) & H_in;
`elsif ALU_BY_LUT

    wire [8:0] CN;
    wire [7:0] F_a;
    wire [7:0] F_l;
    wire OV;
`ifdef ZFLAG_TUNE_1
    wire SZ;
`endif

    assign CN[0] = C_in;

    assign F_l =  (S[1:0] == 0) ? (A & B)
		: (S[1:0] == 1) ? (A ^ B)
		: (S[1:0] == 2) ? (A | B)
		: B;

    assign F = (~S[2]) ? F_a : F_l;

    assign C_out = (~S[2]) ? CN[8] : C_in;
    assign H_out = (~S[2]) ? CN[4] : H_in;
    assign V_out = (~S[2]) ? OV : (S == 7) ? V_in : 1'b0;
`ifdef ZFLAG_TUNE_1
    assign Z_out = (~S[2]) ? SZ & ( F_a[6:0] == 0)
		   : (S == 7) ? Z_in
		   : (F_l[7:0] == 0) ;
`else
     assign Z_out = (S == 7) & Z_in
		  | ( (S == 2) & Z_in | ~((S == 2)|(S == 7)) )
			 & (F[7:0] == 0);
`endif
    wire [1:0] SS = { S[1] , ~S[1] };
    generate
	genvar i;
	for (i = 0; i < 8; i=i + 1)
	    begin : gen_lut
		if (i == 0) assign F_a[i] = 
			  ({ S[1:0] , CN[i] , A[i] , B[i] } ==  0)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  1)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  2)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  3)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  4)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  5)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  6)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  7)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  8)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  9)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 10)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 11)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 12)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 13)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 14)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 15)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 16)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 17)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 18)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 19)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 20)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 21)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 22)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 23)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 24)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 25)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 26)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 27)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 28)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 29)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 30)  ? 1'b1
			:  1'b0;
		else assign F_a[i] = 
			  ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  0)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  1)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  2)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  3)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  4)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  5)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  6)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  7)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  8)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  9)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 10)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 11)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 12)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 13)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 14)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 15)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 16)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 17)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 18)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 19)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 20)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 21)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 22)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 23)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 24)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 25)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 26)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 27)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 28)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 29)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 30)  ? 1'b1
			:  1'b0;

		if (i == 0) assign CN[i+1] = 
			  ({ S[1:0] , CN[i] , A[i] , B[i] } ==  0)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  1)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  2)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  3)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  4)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  5)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  6)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  7)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  8)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } ==  9)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 10)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 11)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 12)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 13)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 14)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 15)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 16)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 17)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 18)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 19)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 20)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 21)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 22)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 23)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 24)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 25)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 26)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 27)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 28)  ? 1'b0
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 29)  ? 1'b1
			: ({ S[1:0] , CN[i] , A[i] , B[i] } == 30)  ? 1'b0
			: 1'b0;
		else assign CN[i+1] = 
			  ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  0)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  1)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  2)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  3)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  4)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  5)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  6)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  7)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  8)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  9)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 10)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 11)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 12)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 13)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 14)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 15)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 16)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 17)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 18)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 19)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 20)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 21)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 22)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 23)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 24)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 25)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 26)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 27)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 28)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 29)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 30)  ? 1'b0
			: 1'b0;

		if (i == 7) assign OV = 
			  ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  0)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  1)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  2)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  3)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  4)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  5)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  6)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  7)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  8)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } ==  9)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 10)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 11)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 12)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 13)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 14)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 15)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 16)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 17)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 18)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 19)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 20)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 21)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 22)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 23)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 24)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 25)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 26)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 27)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 28)  ? 1'b0
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 29)  ? 1'b1
			: ({ SS[1:0] , CN[i] , A[i] , B[i] } == 30)  ? 1'b0
			: 1'b0;

    `ifdef ZFLAG_TUNE_1
		if (i == 7) assign SZ = 
			  ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  0)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  1)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  2)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  3)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  4)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  5)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  6)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  7)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  8)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} ==  9)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 10)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 11)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 12)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 13)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 14)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 15)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 16)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 17)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 18)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 19)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 20)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 21)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 22)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 23)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 24)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 25)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 26)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 27)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 28)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 29)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 30)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 31)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 32)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 33)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 34)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 35)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 36)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 37)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 38)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 39)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 40)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 41)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 42)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 43)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 44)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 45)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 46)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 47)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 48)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 49)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 50)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 51)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 52)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 53)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 54)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 55)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 56)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 57)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 58)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 59)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 60)  ? 1'b0
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 61)  ? 1'b1
			: ({ Z_in, S[1:0], CN[i], A[i], B[i]} == 62)  ? 1'b1
			: 1'b0;
    `endif
	    end
    endgenerate

`endif // ALU_BY_LUT

endmodule

`endif // USE_ALU_MODULE

// 
//
// input: { S[1:0] , C_in , A[0], B[0] } 
// F  01100110 0110100 10110100 101100110
// C  00010001 0001011 10100110 101000100
// 
// OV 00010001 0001100 00100001 001000100
// input: { Z_in, S[1:0] , C_in , A[0], B[0] }
// Z  10010110 1001011 00000000 010010110 10010110 1001011 01001011 010010110
// input: { Z_in, S[1:0] , C_in , A[1:0], B[1:0] }
// Z  1010000010100000101000001010000010100000101000000000000000000000
//    0000000000000000000000000000000010100101101001011010010110100101
//    1010000010100000101000001010000010100000101000000000000000000000
//    1010010110100101000010100000101010100101101001011010010110100101


