/*
 * Copyright (C) 2011 Koji Suzuki
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.1 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

`include "rtavr_defs.v"

`ifdef USE_ALU_MODULE

//`define ALU_Z_TUNE
//`define ALU_Z_TUNE_2

// global defines :  USE_ALU_MODULE , ALU_Z_TUNE_1

//
// Module Name: rtavr_alu
//   ver 001
// 
// Target Device:   xc3s200a-4ft256
// Product Version: ISE 12.4
// Design Goal:     Balanced
//
//                                        all-in-one  
// Number of Slice Flip Flops:              xxx
// Number of 4 input LUTs:                  xxx
// Number of occupied Slices:               xxx
// Total Number of 4 input LUTs:            xxx
// Number of bonded IOBs :                  xxx
//   IOB Flip Flops :                       xxx
// Number of BUFGMUXs :                     xxx
// Number used for Dual Port RAMs           xxx
// Number of RAMB16BWEs :                   xxx
// Clock to Setup 

module rtavr_alu (
		  input [7:0] A
		, input [7:0] B
		, output [7:0] F

		, input C_in
		, input H_in
		, input V_in
		, input Z_in
		, output C_out
		, output H_out
		, output V_out
		, output Z_out

		, input [2:0] S
		);

// S == 0 : ADD (wo carry)
// S == 1 : ADC (with carry)
// S == 2 : SUB (wo carry)
// S == 3 : SBC (with carry)
// S == 4 : AND
// S == 5 : XOR
// S == 6 : OR
// S == 7 : MOV
//
//                      S = 0  1  2  3  4  5  6  7
// C_out  : Carry[8]        o  o  o  o  -  -  -  -
// H_out  : Carry[4]        o  o  o  o  -  -  -  -
// V_out  : Over Flow       o  o  o  o  -  -  -  -
// Z_out  : Zero            o  o  o (o) -  -  -  -
//          (o) if (Z_in == 1)
//
// OP3          --*-:**------------                               HSVNZC
// CP/CPC       000B:01rd:dddd:rrrr	# B=1: CP  B=0: CPC   #   XXXXXX
// SUB/SBC	000B:10rd:dddd:rrrr	# B=1: sub B=0: sbc   #   XXXXXX
// ADD/ADD	000C:11rd:dddd:rrrr	# C=0: add C=1: adc   #   XXXXXX
// AND		0010:00rd:dddd:rrrr	# logical AND         #   -X0XX-
// EOR		0010:01rd:dddd:rrrr	# exclusive OR        #   -X0XX-
// OR		0010:10rd:dddd:rrrr	# inclusive OR        #   -X0XX-
// MOV		0010:11rd:dddd:rrrr	# copy register       #   ------

    wire WC = (S[0] & C_in);
`ifdef ALU_Z_TUNE
`ifdef ALU_Z_TUNE_2
//    (~S[1] & ~WC)  : ADD or ADC(C==0)
//    (~S[1] &  WC)  : ADC(C==1)
//    ( S[1] & ~WC)  : SUB or SBC(C==0)
//    ( S[1] &  WC)  : SBC(C==1)
    wire [8:0] B2  =  S[1] ? ~{1'b0, B} : {1'b0, B};
    wire [8:0] B2I =  B2 + 1;
    wire [8:0] F_a = {1'b0, A} + (( S[1] ^ WC ) ? B2I : B2);
	
    wire [7:0] R0 = (A ^ B);
    wire [7:0] R1 = (A ^ (B-1));
    wire [7:0] R2 = (A ^ B2I);

`else  // ALU_Z_TUNE_2
    wire [8:0] F_a = 
	(~S[1]) ? ({1'b0, A}  + {1'b0, B } + {8'b0, WC })
		: ({1'b0, A}  - {1'b0, B } - {8'b0, WC });

    wire [7:0] R0 = (A ^ B);
    wire [7:0] R1 = (A ^ (B-1));
    wire [7:0] R2 = (A ^ (~B+1));

`endif  // ALU_Z_TUNE_2

    wire F_a_z =  (~S[1] &  WC) ? (R0 == 8'hff) 
		: (~S[1] & ~WC) ? (R1 == 8'hff) 
		: ( S[1] &  WC) ? (R2 == 8'hff) 
		: (R0 == 8'h00) ;

    wire [7:0] F_l =  (S[1:0] == 0) ? (A & B)
		: (S[1:0] == 1) ? (A ^ B)
		: (S[1:0] == 2) ? (A | B)
		: B;

    assign F = (~S[2]) ? F_a[7:0] : F_l[7:0];

    assign C_out = (~S[2]) ? F_a[8] : C_in; 

`ifdef ZZZ
    assign Z_out = (~S[2]) ? (Z_in | (S[1:0] != 3)) & F_a_z
		  : (S[1:0] == 3) & Z_in | (S[1:0] != 3) & (F_l == 0);
`else
    assign Z_out = (S == 7) ? Z_in 
		  : (S == 3) ? ( Z_in & F_a_z )
		  : (~S[2]) ? F_a_z
		  : (F_l == 0);
`endif

`else // ALU_Z_TUNE
//`define ZZ2
`ifdef ZZ2
     wire [8:0] ALU_OUT = ((S == 0) | (S == 1)) ? ({1'b0, A}
				 + {1'b0, B } + {8'b0, WC })
		     : ((S == 2) | (S == 3)) ? ({1'b0,A}
				 - {1'b0, B } - {8'b0, WC })
		     : (S == 4) ? {C_in, (A & B) }
		     : (S == 5) ? {C_in, (A ^ B) }
		     : (S == 6) ?  {C_in, (A | B) }
		     : {C_in, B } ;
`else
     // same but fast ?
     wire [8:0] ALU_OUT = ((S == 0)|(S == 1))
			   ? ({1'b0, A} + ({1'b0, B } + {8'b0, WC }))
		     : ((S == 2)|(S == 3))
			   ? ({1'b0, A} - ({1'b0, B } + {8'b0, WC }))
		     : (S == 4) ? {C_in, (A & B)}
		     : (S == 5) ? {C_in, (A ^ B)}
		     : (S == 6) ? {C_in, (A | B)}
		     : {C_in, B } ;
`endif
     assign F = ALU_OUT[7:0];

     assign C_out = ALU_OUT[8]; 
//`define ZZ3
`ifdef ZZ3
     assign Z_out = (S == 7) & Z_in
		  | ( (S == 3) & Z_in | ~((S == 3)|(S == 7)) )
			 & (ALU_OUT[7:0] == 0);
`else
     // -1 slices
     assign Z_out = (S == 7) ?  Z_in
		  : (S == 3) ? (Z_in & (ALU_OUT[7:0] == 0))
		  :  (ALU_OUT[7:0] == 0);
`endif

`endif  // ALU_Z_TUNE

     // common part
`define ZZ4
`ifdef ZZ4
     assign V_out = ((S == 0)|(S == 1))
			& (F[7] & ~A[7] & ~B[7] | ~F[7] &  A[7] &  B[7] )
	      | ((S == 2)|(S == 3))
			& (F[7] & ~A[7] &  B[7] | ~F[7] &  A[7] & ~B[7] )
	      | (S == 7) & V_in ;
	      //  ((S == 4)|(S == 5)|(S == 6)) ? 0


`else
     // same slices
     assign V_out = ((S == 0)|(S == 1)) 
			? (F[7] & ~A[7] & ~B[7] | ~F[7] &  A[7] &  B[7] )
	      : ((S == 2)|(S == 3))
			? (F[7] & ~A[7] &  B[7] | ~F[7] &  A[7] & ~B[7] )
	      : (S == 7) ? V_in
	      : 0;
`endif
`define ZZ5
`ifdef ZZ5
     assign H_out = ((S == 0) | (S == 1))
			 & ( ~F[3] & A[3] | ~F[3] & B[3] |  A[3] & B[3] )
	       | ((S == 2) | (S == 3))
			 & ( F[3] & ~A[3] |  F[3] & B[3] | ~A[3] & B[3] )
	       | (S[2]) & H_in;
`else
     // + 4 slices
     assign H_out = ((S == 0) | (S == 1))
			 ? ( ~F[3] & A[3] | ~F[3] & B[3] |  A[3] & B[3] )
	       : ((S == 2) | (S == 3))
			 ? ( F[3] & ~A[3] |  F[3] & B[3] | ~A[3] & B[3] )
	       : (S[2]) & H_in;
`endif


endmodule

`endif // USE_ALU_MODULE
